@@ -772,21 +772,25 @@ static int xmlpost_initial_req(struct openconnect_info *vpninfo,
 {
 	xmlNodePtr root, node;
 	xmlDocPtr doc = xmlpost_new_query(vpninfo, "init", &root);
-	char *url;
-	int result;
+	struct oc_text_buf *url_buf;
 
 	if (!doc)
 		return -ENOMEM;
 
+	url_buf = buf_alloc();
+	buf_append(url_buf, "https://%s", vpninfo->hostname);
+	if (vpninfo->port != 443)
+		buf_append(url_buf, ":%d", vpninfo->port);
+	/* Do we *need* to omit the trailing / here when no path? */
 	if (vpninfo->urlpath)
-		result = asprintf(&url, "https://%s/%s", vpninfo->hostname, vpninfo->urlpath);
-	else
-		result = asprintf(&url, "https://%s", vpninfo->hostname);
+		buf_append(url_buf, "/%s", vpninfo->urlpath);
 
-	if (result == -1)
+	if (buf_error(url_buf)) {
+		buf_free(url_buf);
 		goto bad;
-	node = xmlNewTextChild(root, NULL, XCAST("group-access"), XCAST(url));
-	free(url);
+	}
+	node = xmlNewTextChild(root, NULL, XCAST("group-access"), XCAST(url_buf->data));
+	buf_free(url_buf);
 	if (!node)
 		goto bad;
 	if (cert_fail) {
@@ -802,6 +806,7 @@ static int xmlpost_initial_req(struct openconnect_info *vpninfo,
 	return xmlpost_complete(doc, request_body);
 
 bad:
+	buf_free(url_buf);
 	xmlpost_complete(doc, NULL);
 	return -ENOMEM;
 }
@@ -919,7 +924,11 @@ static int fetch_config(struct openconnect_info *vpninfo)
 	}
 
 	buf = buf_alloc();
-	buf_append(buf, "GET %s HTTP/1.1\r\n", vpninfo->profile_url);
+
+	if (vpninfo->port != 443)
+		buf_append(buf, "GET %s:%d HTTP/1.1\r\n", vpninfo->profile_url, vpninfo->port);
+	else
+		buf_append(buf, "GET %s HTTP/1.1\r\n", vpninfo->profile_url);
 	cstp_common_headers(vpninfo, buf);
 	if (vpninfo->xmlpost)
 		buf_append(buf, "Cookie: webvpn=%s\r\n", vpninfo->cookie);
