@@ -30,6 +30,14 @@
 #include <stdarg.h>
 #ifdef HAVE_LZ4
 #include <lz4.h>
+#ifndef HAVE_LZ4_COMPRESS_DEFAULT
+#define LZ4_compress_default LZ4_compress_limitedOutput
+#endif
+#endif
+
+#if defined(__linux__)
+/* For TCP_INFO */
+# include <linux/tcp.h>
 #endif
 
 #include "openconnect-internal.h"
@@ -68,9 +76,13 @@ static const struct pkt dpd_resp_pkt = {
 	{ .cstp.hdr = { 'S', 'T', 'F', 1, 0, 0, AC_PKT_DPD_RESP, 0 } }
 };
 
+#define UDP_HEADER_SIZE 8
+#define IPV4_HEADER_SIZE 20
+#define IPV6_HEADER_SIZE 40
+
 /* Calculate MTU to request. Old servers simply use the X-CSTP-MTU: header,
  * which represents the tunnel MTU, while new servers do calculations on the
- * X-CSTP-Base-MTU: header which represents the cleartext MTU between client
+ * X-CSTP-Base-MTU: header which represents the link MTU between client
  * and server.
  *
  * If possible, the legacy MTU value should be the TCP MSS less 5 bytes of
@@ -99,34 +111,47 @@ static void calculate_mtu(struct openconnect_info *vpninfo, int *base_mtu, int *
 			vpn_progress(vpninfo, PRG_DEBUG,
 				     _("TCP_INFO rcv mss %d, snd mss %d, adv mss %d, pmtu %d\n"),
 				     ti.tcpi_rcv_mss, ti.tcpi_snd_mss, ti.tcpi_advmss, ti.tcpi_pmtu);
-			if (!*base_mtu)
+
+			if (!*base_mtu) {
 				*base_mtu = ti.tcpi_pmtu;
-			if (!*mtu) {
+			}
+
+			if (!*base_mtu) {
 				if (ti.tcpi_rcv_mss < ti.tcpi_snd_mss)
-					*mtu = ti.tcpi_rcv_mss - 13;
+					*base_mtu = ti.tcpi_rcv_mss - 13;
 				else
-					*mtu = ti.tcpi_snd_mss - 13;
+					*base_mtu = ti.tcpi_snd_mss - 13;
 			}
 		}
 	}
 #endif
 #ifdef TCP_MAXSEG
-	if (!*mtu) {
+	if (!*base_mtu) {
 		int mss;
 		socklen_t mss_size = sizeof(mss);
 		if (!getsockopt(vpninfo->ssl_fd, IPPROTO_TCP, TCP_MAXSEG,
 				&mss, &mss_size)) {
 			vpn_progress(vpninfo, PRG_DEBUG, _("TCP_MAXSEG %d\n"), mss);
-			*mtu = mss - 13;
+			*base_mtu = mss - 13;
 		}
 	}
 #endif
-	if (!*mtu) {
+	if (!*base_mtu) {
 		/* Default */
-		*mtu = 1406;
+		*base_mtu = 1406;
+	}
+
+	if (*base_mtu < 1280)
+		*base_mtu = 1280;
+
+	if (!*mtu) {
+		/* remove IP/UDP and DTLS overhead from base MTU to calculate tunnel MTU */
+		*mtu = *base_mtu - DTLS_OVERHEAD - UDP_HEADER_SIZE;
+		if (vpninfo->peer_addr->sa_family == AF_INET6)
+			*mtu -= IPV6_HEADER_SIZE;
+		else
+			*mtu -= IPV4_HEADER_SIZE;
 	}
-	if (*mtu < 1280)
-		*mtu = 1280;
 }
 
 static void append_compr_types(struct oc_text_buf *buf, const char *proto, int avail)
@@ -166,6 +191,26 @@ static void append_mobile_headers(struct openconnect_info *vpninfo, struct oc_te
 	}
 }
 
+static int parse_hex_val(const char *str, unsigned char *storage, unsigned int max_storage_len, int *changed)
+{
+	int len = strlen(str);
+	unsigned i;
+
+	if (len % 2 == 1 || len > 2*max_storage_len) {
+		return -EINVAL;
+	}
+
+	for (i = 0; i < len; i += 2) {
+		unsigned char c = unhex(str + i);
+		if (storage[i/2] != c) {
+			storage[i/2] = c;
+			*changed = 1;
+		}
+	}
+
+	return len/2;
+}
+
 static int start_cstp_connection(struct openconnect_info *vpninfo)
 {
 	struct oc_text_buf *reqbuf;
@@ -181,7 +226,7 @@ static int start_cstp_connection(struct openconnect_info *vpninfo)
 	const char *old_netmask = vpninfo->ip_info.netmask;
 	const char *old_addr6 = vpninfo->ip_info.addr6;
 	const char *old_netmask6 = vpninfo->ip_info.netmask6;
-	int base_mtu, mtu;
+	int base_mtu = 0, mtu = 0;
 
 	/* Clear old options which will be overwritten */
 	vpninfo->ip_info.addr = vpninfo->ip_info.netmask = NULL;
@@ -196,10 +241,14 @@ static int start_cstp_connection(struct openconnect_info *vpninfo)
 
  retry:
 	calculate_mtu(vpninfo, &base_mtu, &mtu);
+	vpninfo->cstp_basemtu = base_mtu;
 
 	reqbuf = buf_alloc();
 	buf_append(reqbuf, "CONNECT /CSCOSSLC/tunnel HTTP/1.1\r\n");
-	buf_append(reqbuf, "Host: %s\r\n", vpninfo->hostname);
+	if (vpninfo->port != 443)
+		buf_append(reqbuf, "Host: %s:%d\r\n", vpninfo->hostname, vpninfo->port);
+	else
+		buf_append(reqbuf, "Host: %s\r\n", vpninfo->hostname);
 	buf_append(reqbuf, "User-Agent: %s\r\n", vpninfo->useragent);
 	buf_append(reqbuf, "Cookie: webvpn=%s\r\n", vpninfo->cookie);
 	buf_append(reqbuf, "X-CSTP-Version: 1\r\n");
@@ -208,15 +257,18 @@ static int start_cstp_connection(struct openconnect_info *vpninfo)
 	append_mobile_headers(vpninfo, reqbuf);
 	append_compr_types(reqbuf, "CSTP", vpninfo->req_compr);
 
-	if (base_mtu)
-		buf_append(reqbuf, "X-CSTP-Base-MTU: %d\r\n", base_mtu);
-	buf_append(reqbuf, "X-CSTP-MTU: %d\r\n", mtu);
+	buf_append(reqbuf, "X-CSTP-Base-MTU: %d\r\n", base_mtu);
+	if (mtu)
+		buf_append(reqbuf, "X-CSTP-MTU: %d\r\n", mtu);
 	buf_append(reqbuf, "X-CSTP-Address-Type: %s\r\n",
 			       vpninfo->disable_ipv6 ? "IPv4" : "IPv6,IPv4");
 	if (!vpninfo->disable_ipv6)
 		buf_append(reqbuf, "X-CSTP-Full-IPv6-Capability: true\r\n");
 #ifdef HAVE_DTLS
 	if (vpninfo->dtls_state != DTLS_DISABLED) {
+		/* The X-DTLS-Master-Secret is only used for the legacy protocol negotation
+		 * which required the client to send explicitly the secret. In the PSK-NEGOTIATE
+		 * method, the master secret is implicitly agreed on */
 		buf_append(reqbuf, "X-DTLS-Master-Secret: ");
 		for (i = 0; i < sizeof(vpninfo->dtls_secret); i++) {
 			buf_append(reqbuf, "%02X", vpninfo->dtls_secret[i]);
@@ -356,25 +408,39 @@ static int start_cstp_connection(struct openconnect_info *vpninfo)
 					mtu = dtlsmtu;
 			} else if (!strcmp(buf + 7, "Session-ID")) {
 				int dtls_sessid_changed = 0;
+				int vsize;
 
-				if (strlen(colon) != 64) {
+				vsize = parse_hex_val(colon, vpninfo->dtls_session_id, sizeof(vpninfo->dtls_session_id), &dtls_sessid_changed);
+				if (vsize != 32) {
 					vpn_progress(vpninfo, PRG_ERR,
-						     _("X-DTLS-Session-ID not 64 characters; is: \"%s\"\n"),
-						     colon);
+					     _("X-DTLS-Session-ID not 64 characters; is: \"%s\"\n"),
+					     	colon);
 					vpninfo->dtls_attempt_period = 0;
 					return -EINVAL;
 				}
-				for (i = 0; i < 64; i += 2) {
-					unsigned char c = unhex(colon + i);
-					if (vpninfo->dtls_session_id[i/2] != c) {
-						vpninfo->dtls_session_id[i/2] = c;
-						dtls_sessid_changed = 1;
-					}
-				}
+
 				sessid_found = 1;
 
 				if (dtls_sessid_changed && vpninfo->dtls_state > DTLS_SLEEPING)
 					vpninfo->dtls_need_reconnect = 1;
+			} else if (!strcmp(buf + 7, "App-ID")) {
+				int dtls_appid_changed = 0;
+				int vsize;
+
+				vsize = parse_hex_val(colon, vpninfo->dtls_app_id, sizeof(vpninfo->dtls_app_id), &dtls_appid_changed);
+				if (vsize <= 0) {
+					vpn_progress(vpninfo, PRG_ERR,
+					     _("X-DTLS-Session-ID is invalid; is: \"%s\"\n"),
+					     	colon);
+					vpninfo->dtls_attempt_period = 0;
+					return -EINVAL;
+				}
+
+				vpninfo->dtls_app_id_size = vsize;
+				sessid_found = 1;
+
+				if (dtls_appid_changed && vpninfo->dtls_state > DTLS_SLEEPING)
+					vpninfo->dtls_need_reconnect = 1;
 			} else if (!strcmp(buf + 7, "Content-Encoding")) {
 				if (!strcmp(colon, "lzs"))
 					vpninfo->dtls_compr = COMPR_LZS;
@@ -422,6 +488,8 @@ static int start_cstp_connection(struct openconnect_info *vpninfo)
 					     colon);
 				return -EINVAL;
 			}
+		} else if (!strcmp(buf + 7, "Base-MTU")) {
+			vpninfo->cstp_basemtu = atol(colon);
 		} else if (!strcmp(buf + 7, "MTU")) {
 			int cstpmtu = atol(colon);
 			if (cstpmtu > mtu)
@@ -794,8 +862,8 @@ int compress_packet(struct openconnect_info *vpninfo, int compr_type, struct pkt
 		if (this->len < 40)
 			return -EFBIG;
 
-		ret = LZ4_compress_limitedOutput((void*)this->data, (void*)vpninfo->deflate_pkt->data, this->len,
-				   this->len);
+		ret = LZ4_compress_default((void*)this->data, (void*)vpninfo->deflate_pkt->data,
+					   this->len, this->len);
 		if (ret <= 0) {
 			if (ret == 0)
 				ret = -EFBIG;
