@@ -99,26 +99,26 @@ static P11KitPin *p11kit_pin_callback(const char *pin_source, P11KitUri *pin_uri
 	 gnutls_check_version(#a "." #b "." #c)))
 
 /* Helper functions for reading/writing lines over SSL. */
-static int openconnect_gnutls_write(struct openconnect_info *vpninfo, char *buf, size_t len)
+static int _openconnect_gnutls_write(gnutls_session_t ses, int fd, struct openconnect_info *vpninfo, char *buf, size_t len)
 {
 	size_t orig_len = len;
 
 	while (len) {
-		int done = gnutls_record_send(vpninfo->https_sess, buf, len);
+		int done = gnutls_record_send(ses, buf, len);
 		if (done > 0)
 			len -= done;
-		else if (done == GNUTLS_E_AGAIN) {
+		else if (done == GNUTLS_E_AGAIN || done == GNUTLS_E_INTERRUPTED) {
 			/* Wait for something to happen on the socket, or on cmd_fd */
 			fd_set wr_set, rd_set;
-			int maxfd = vpninfo->ssl_fd;
+			int maxfd = fd;
 
 			FD_ZERO(&wr_set);
 			FD_ZERO(&rd_set);
 
-			if (gnutls_record_get_direction(vpninfo->https_sess))
-				FD_SET(vpninfo->ssl_fd, &wr_set);
+			if (gnutls_record_get_direction(ses))
+				FD_SET(fd, &wr_set);
 			else
-				FD_SET(vpninfo->ssl_fd, &rd_set);
+				FD_SET(fd, &rd_set);
 
 			cmd_fd_set(vpninfo, &rd_set, &maxfd);
 			select(maxfd + 1, &rd_set, &wr_set, NULL, NULL);
@@ -135,48 +135,93 @@ static int openconnect_gnutls_write(struct openconnect_info *vpninfo, char *buf,
 	return orig_len;
 }
 
-static int openconnect_gnutls_read(struct openconnect_info *vpninfo, char *buf, size_t len)
+static int openconnect_gnutls_write(struct openconnect_info *vpninfo, char *buf, size_t len)
 {
-	int done;
+	return _openconnect_gnutls_write(vpninfo->https_sess, vpninfo->ssl_fd, vpninfo, buf, len);
+}
 
-	while ((done = gnutls_record_recv(vpninfo->https_sess, buf, len)) < 0) {
-		if (done == GNUTLS_E_AGAIN) {
+int openconnect_dtls_write(struct openconnect_info *vpninfo, void *buf, size_t len)
+{
+	return _openconnect_gnutls_write(vpninfo->dtls_ssl, vpninfo->dtls_fd, vpninfo, buf, len);
+}
+
+static int _openconnect_gnutls_read(gnutls_session_t ses, int fd, struct openconnect_info *vpninfo, char *buf, size_t len, unsigned ms)
+{
+	int done, ret;
+	struct timeval timeout, *tv = NULL;
+
+	if (ms) {
+		timeout.tv_sec = ms/1000;
+		timeout.tv_usec = (ms%1000)*1000;
+		tv = &timeout;
+	}
+
+	while ((done = gnutls_record_recv(ses, buf, len)) < 0) {
+		if (done == GNUTLS_E_AGAIN || done == GNUTLS_E_INTERRUPTED) {
 			/* Wait for something to happen on the socket, or on cmd_fd */
 			fd_set wr_set, rd_set;
-			int maxfd = vpninfo->ssl_fd;
+			int maxfd = fd;
 
 			FD_ZERO(&wr_set);
 			FD_ZERO(&rd_set);
 
-			if (gnutls_record_get_direction(vpninfo->https_sess))
-				FD_SET(vpninfo->ssl_fd, &wr_set);
+			if (gnutls_record_get_direction(ses))
+				FD_SET(fd, &wr_set);
 			else
-				FD_SET(vpninfo->ssl_fd, &rd_set);
+				FD_SET(fd, &rd_set);
 
 			cmd_fd_set(vpninfo, &rd_set, &maxfd);
-			select(maxfd + 1, &rd_set, &wr_set, NULL, NULL);
+			ret = select(maxfd + 1, &rd_set, &wr_set, NULL, tv);
 			if (is_cancel_pending(vpninfo, &rd_set)) {
 				vpn_progress(vpninfo, PRG_ERR, _("SSL read cancelled\n"));
-				return -EINTR;
+				done = -EINTR;
+				goto cleanup;
+			}
+
+			if (ret == 0) {
+				done = -ETIMEDOUT;
+				goto cleanup;
 			}
 		} else if (done == GNUTLS_E_PREMATURE_TERMINATION) {
 			/* We've seen this with HTTP 1.0 responses followed by abrupt
 			   socket closure and no clean SSL shutdown.
 			   https://bugs.launchpad.net/bugs/1225276 */
 			vpn_progress(vpninfo, PRG_DEBUG, _("SSL socket closed uncleanly\n"));
-			return 0;
+			done = 0;
+			goto cleanup;
 		} else if (done == GNUTLS_E_REHANDSHAKE) {
 			int ret = cstp_handshake(vpninfo, 0);
-			if (ret)
-				return ret;
+			if (ret) {
+				done = ret;
+				goto cleanup;
+			}
 		} else {
 			vpn_progress(vpninfo, PRG_ERR, _("Failed to read from SSL socket: %s\n"),
 				     gnutls_strerror(done));
-			return -EIO;
+			if (done == GNUTLS_E_TIMEDOUT) {
+				done = -ETIMEDOUT;
+				goto cleanup;
+			} else {
+				done = -EIO;
+				goto cleanup;
+			}
 		}
 
 	}
+
+ cleanup:
 	return done;
+
+}
+
+static int openconnect_gnutls_read(struct openconnect_info *vpninfo, char *buf, size_t len)
+{
+	return _openconnect_gnutls_read(vpninfo->https_sess, vpninfo->ssl_fd, vpninfo, buf, len, 0);
+}
+
+int openconnect_dtls_read(struct openconnect_info *vpninfo, void *buf, size_t len, unsigned ms)
+{
+	return _openconnect_gnutls_read(vpninfo->dtls_ssl, vpninfo->dtls_fd, vpninfo, buf, len, ms);
 }
 
 static int openconnect_gnutls_gets(struct openconnect_info *vpninfo, char *buf, size_t len)
@@ -204,7 +249,7 @@ static int openconnect_gnutls_gets(struct openconnect_info *vpninfo, char *buf,
 				buf[i] = 0;
 				return i;
 			}
-		} else if (ret == GNUTLS_E_AGAIN) {
+		} else if (ret == GNUTLS_E_AGAIN || ret == GNUTLS_E_INTERRUPTED) {
 			/* Wait for something to happen on the socket, or on cmd_fd */
 			fd_set rd_set, wr_set;
 			int maxfd = vpninfo->ssl_fd;
@@ -247,7 +292,7 @@ int ssl_nonblock_read(struct openconnect_info *vpninfo, void *buf, int maxlen)
 	if (ret > 0)
 		return ret;
 
-	if (ret != GNUTLS_E_AGAIN) {
+	if (ret != GNUTLS_E_AGAIN && ret != GNUTLS_E_INTERRUPTED) {
 		vpn_progress(vpninfo, PRG_ERR,
 			     _("SSL read error: %s; reconnecting.\n"),
 			     gnutls_strerror(ret));
@@ -264,7 +309,7 @@ int ssl_nonblock_write(struct openconnect_info *vpninfo, void *buf, int buflen)
 	if (ret > 0)
 		return ret;
 
-	if (ret == GNUTLS_E_AGAIN) {
+	if (ret == GNUTLS_E_AGAIN || ret == GNUTLS_E_INTERRUPTED) {
 		/*
 		 * Before 3.3.13, GnuTLS could return zero instead of one,
 		 * indicating that it was waiting for a read when in fact
@@ -921,8 +966,8 @@ static int import_openssl_pem(struct openconnect_info *vpninfo,
 			}
 			keylen += ofs;
 
-			/* If there appears to be more padding than required, fail */
-			if (b64_data.size - keylen >= blocksize)
+			/* If there appears to be more or less padding than required, fail */
+			if (b64_data.size - keylen > blocksize || b64_data.size < keylen + 1)
 				goto fail;
 
 			/* If the padding bytes aren't all equal to the amount of padding, fail */
@@ -1490,12 +1535,39 @@ static int load_certificate(struct openconnect_info *vpninfo)
 		}
 		free(pass);
 		vpninfo->cert_password = NULL;
+	} else if (!gnutls_x509_privkey_import(key, &fdata, GNUTLS_X509_FMT_DER) ||
+		   !gnutls_x509_privkey_import_pkcs8(key, &fdata, GNUTLS_X509_FMT_DER,
+						     NULL, GNUTLS_PKCS_PLAIN)) {
+		/* Unencrypted DER (PKCS#1 or PKCS#8) */
 	} else {
-		vpn_progress(vpninfo, PRG_ERR,
-			     _("Failed to determine type of private key %s\n"),
-			     vpninfo->sslkey);
-		ret = -EINVAL;
-		goto out;
+		/* Last chance: try encrypted PKCS#8 DER. And give up if it's not that */
+		char *pass = vpninfo->cert_password;
+
+		while ((err = gnutls_x509_privkey_import_pkcs8(key, &fdata,
+							       GNUTLS_X509_FMT_DER,
+							       pass?:"", 0))) {
+			if (err != GNUTLS_E_DECRYPTION_FAILED) {
+				vpn_progress(vpninfo, PRG_ERR,
+					     _("Failed to determine type of private key %s\n"),
+					     vpninfo->sslkey);
+				ret = -EINVAL;
+				goto out;
+			}
+			vpninfo->cert_password = NULL;
+			if (pass) {
+				vpn_progress(vpninfo, PRG_ERR,
+					     _("Failed to decrypt PKCS#8 certificate file\n"));
+				free(pass);
+			}
+			err = request_passphrase(vpninfo, "openconnect_pem",
+						 &pass, _("Enter PKCS#8 pass phrase:"));
+			if (err) {
+				ret = -EINVAL;
+				goto out;
+			}
+		}
+		free(pass);
+		vpninfo->cert_password = NULL;
 	}
 
 	/* Now attempt to make sure we use the *correct* certificate, to match
@@ -1696,8 +1768,8 @@ static int load_certificate(struct openconnect_info *vpninfo)
 					gnutls_free(t.data);
 				}
 				if (err) {
-					vpn_progress(vpninfo, PRG_ERR,
-						     "Got no issuer from PKCS#11\n");
+					vpn_progress(vpninfo, PRG_TRACE,
+						     _("Got no issuer from PKCS#11\n"));
 				} else {
 					get_cert_name(issuer, name, sizeof(name));
 
@@ -1855,11 +1927,11 @@ int get_cert_md5_fingerprint(struct openconnect_info *vpninfo,
 
 static int set_peer_cert_hash(struct openconnect_info *vpninfo)
 {
-	unsigned char sha1[SHA1_SIZE];
+	unsigned char hash[SHA256_SIZE];
 	size_t shalen;
 	gnutls_pubkey_t pkey;
 	gnutls_datum_t d;
-	int i, err;
+	int err;
 
 	err = gnutls_pubkey_init(&pkey);
 	if (err)
@@ -1897,9 +1969,18 @@ static int set_peer_cert_hash(struct openconnect_info *vpninfo)
 	}
 #endif
 	gnutls_pubkey_deinit(pkey);
-	shalen = SHA1_SIZE;
+	shalen = SHA256_SIZE;
 
-	err = gnutls_fingerprint(GNUTLS_DIG_SHA1, &d, sha1, &shalen);
+	err = gnutls_fingerprint(GNUTLS_DIG_SHA256, &d, hash, &shalen);
+	if (err) {
+		gnutls_free(d.data);
+		return err;
+	}
+
+	vpninfo->peer_cert_sha256 = openconnect_bin2hex("sha256:", hash, shalen);
+
+	shalen = SHA1_SIZE;
+	err = gnutls_fingerprint(GNUTLS_DIG_SHA1, &d, hash, &shalen);
 	if (err) {
 		gnutls_free(d.data);
 		return err;
@@ -1907,12 +1988,7 @@ static int set_peer_cert_hash(struct openconnect_info *vpninfo)
 
 	gnutls_free(d.data);
 
-	vpninfo->peer_cert_hash = malloc(SHA1_SIZE * 2 + 6);
-	if (vpninfo->peer_cert_hash) {
-		snprintf(vpninfo->peer_cert_hash, 6, "sha1:");
-		for (i = 0; i < shalen; i++)
-			sprintf(&vpninfo->peer_cert_hash[i*2 + 5], "%02x", sha1[i]);
-	}
+	vpninfo->peer_cert_sha1 = openconnect_bin2hex("sha1:", hash, shalen);
 
 	return 0;
 }
@@ -2145,8 +2221,10 @@ int openconnect_open_https(struct openconnect_info *vpninfo)
 		gnutls_x509_crt_deinit(vpninfo->peer_cert);
 		vpninfo->peer_cert = NULL;
 	}
-	free(vpninfo->peer_cert_hash);
-	vpninfo->peer_cert_hash = NULL;
+	free(vpninfo->peer_cert_sha1);
+	vpninfo->peer_cert_sha1 = NULL;
+	free(vpninfo->peer_cert_sha256);
+	vpninfo->peer_cert_sha256 = NULL;
 	gnutls_free(vpninfo->cstp_cipher);
 	vpninfo->cstp_cipher = NULL;
 
@@ -2354,7 +2432,7 @@ int cstp_handshake(struct openconnect_info *vpninfo, unsigned init)
 	ssl_sock = (intptr_t)gnutls_transport_get_ptr(vpninfo->https_sess);
 
 	while ((err = gnutls_handshake(vpninfo->https_sess))) {
-		if (err == GNUTLS_E_AGAIN) {
+		if (err == GNUTLS_E_AGAIN || err == GNUTLS_E_INTERRUPTED) {
 			fd_set rd_set, wr_set;
 			int maxfd = ssl_sock;
 
@@ -2375,7 +2453,7 @@ int cstp_handshake(struct openconnect_info *vpninfo, unsigned init)
 				closesocket(ssl_sock);
 				return -EINTR;
 			}
-		} else if (err == GNUTLS_E_INTERRUPTED || gnutls_error_is_fatal(err)) {
+		} else if (gnutls_error_is_fatal(err)) {
 			vpn_progress(vpninfo, PRG_ERR, _("SSL connection failure: %s\n"),
 							 gnutls_strerror(err));
 			gnutls_deinit(vpninfo->https_sess);
@@ -2467,6 +2545,117 @@ void openconnect_close_https(struct openconnect_info *vpninfo, int final)
 	}
 }
 
+#if GNUTLS_VERSION_NUMBER >= 0x030400
+static int ext_recv_client(gnutls_session_t sess, const unsigned char *buf, size_t buflen)
+{
+	/* we shouldn't have received that */
+	return 0;
+}
+
+static int ext_send_client(gnutls_session_t sess, gnutls_buffer_t extdata)
+{
+	struct openconnect_info *vpninfo = gnutls_session_get_ptr(sess);
+
+	if (vpninfo->dtls_ssl != sess)
+		return 0;
+
+	if (vpninfo->dtls_app_id_size > 0) {
+		uint8_t size = vpninfo->dtls_app_id_size;
+		int ret;
+
+		ret = gnutls_buffer_append_data(extdata, &size, 1);
+		if (ret < 0)
+			return ret;
+
+		ret = gnutls_buffer_append_data(extdata, vpninfo->dtls_app_id, vpninfo->dtls_app_id_size);
+		if (ret < 0)
+			return ret;
+
+		return vpninfo->dtls_app_id_size + 1;
+	}
+
+	return 0;
+}
+#else
+
+/* previously to 3.4.0 we can only use internal-but-exported APIs
+ */
+typedef int (*gnutls_ext_recv_func) (gnutls_session_t session,
+				     const unsigned char *data,
+				     size_t len);
+typedef int (*gnutls_ext_send_func) (gnutls_session_t session,
+				     void* extdata);
+int _gnutls_buffer_append_data(void *, const void *data, size_t data_size);
+
+typedef struct {
+	const char *name;
+	uint16_t type;
+	int parse_type;
+
+	/* this function must return 0 when Not Applicable
+	 * size of extension data if ok
+	 * < 0 on other error.
+	 */
+	gnutls_ext_recv_func recv_func;
+
+	/* this function must return 0 when Not Applicable
+	 * size of extension data if ok
+	 * GNUTLS_E_INT_RET_0 if extension data size is zero
+	 * < 0 on other error.
+	 */
+	gnutls_ext_send_func send_func;
+
+	void *deinit_func;	/* this will be called to deinitialize
+							 * internal data 
+							 */
+	void *pack_func;	/* packs internal data to machine independent format */
+	void *unpack_func;	/* unpacks internal data */
+	void *epoch_func;	/* called after the handshake is finished */
+} extension_entry_st;
+
+int _gnutls_ext_register(extension_entry_st *);
+
+static int ext_recv_client(gnutls_session_t sess, const unsigned char *buf, size_t buflen)
+{
+	/* we shouldn't have received that */
+	return 0;
+}
+static int ext_send_client(gnutls_session_t sess, void *extdata)
+{
+	struct openconnect_info *vpninfo = gnutls_session_get_ptr(sess);
+
+	if (vpninfo->dtls_ssl != sess)
+		return 0;
+
+	if (vpninfo->dtls_app_id_size > 0) {
+		uint8_t size = vpninfo->dtls_app_id_size;
+		int ret;
+
+		ret = _gnutls_buffer_append_data(extdata, &size, 1);
+		if (ret < 0)
+			return ret;
+		ret = _gnutls_buffer_append_data(extdata, vpninfo->dtls_app_id, vpninfo->dtls_app_id_size);
+		if (ret < 0)
+			return ret;
+
+		return vpninfo->dtls_app_id_size + 1;
+	}
+
+	return 0;
+}
+
+extension_entry_st ext_app_id  = {
+	.name = "app-id",
+	.type = DTLS_APP_ID_EXT,
+	.parse_type = 2,
+	.recv_func = ext_recv_client,
+	.send_func = ext_send_client,
+	.pack_func = NULL,
+	.unpack_func = NULL,
+	.deinit_func = NULL
+};
+#endif
+
 int openconnect_init_ssl(void)
 {
 #ifdef _WIN32
@@ -2477,6 +2666,12 @@ int openconnect_init_ssl(void)
 	if (gnutls_global_init())
 		return -EIO;
 
+#if GNUTLS_VERSION_NUMBER >= 0x030400
+	gnutls_ext_register("APP-ID", DTLS_APP_ID_EXT, GNUTLS_EXT_TLS, ext_recv_client, ext_send_client, NULL, NULL, NULL);
+#else
+	_gnutls_ext_register(&ext_app_id);
+#endif
+
 	return 0;
 }
 
@@ -2506,6 +2701,19 @@ int openconnect_sha1(unsigned char *result, void *data, int datalen)
 	return 0;
 }
 
+int openconnect_sha256(unsigned char *result, void *data, int datalen)
+{
+	gnutls_datum_t d;
+	size_t shalen = SHA256_SIZE;
+
+	d.data = data;
+	d.size = datalen;
+	if (gnutls_fingerprint(GNUTLS_DIG_SHA256, &d, result, &shalen))
+		return -1;
+
+	return 0;
+}
+
 int openconnect_md5(unsigned char *result, void *data, int datalen)
 {
 	gnutls_datum_t d;
@@ -2572,6 +2780,13 @@ static int gnutls_pin_callback(void *priv, int attempt, const char *uri,
 		(*cache)->token = strdup(uri);
 	}
 
+	if (!attempt && vpninfo->cert_password) {
+		snprintf(pin, pin_max, "%s", vpninfo->cert_password);
+		(*cache)->pin = vpninfo->cert_password;
+		vpninfo->cert_password = NULL;
+		return 0;
+	}
+
 	memset(&f, 0, sizeof(f));
 	f.auth_id = (char *)"pkcs11_pin";
 	f.opts = &o;
