@@ -41,11 +41,11 @@ struct oc_text_buf *buf_alloc(void)
 	return calloc(1, sizeof(struct oc_text_buf));
 }
 
-void buf_append_urlencoded(struct oc_text_buf *buf, char *str)
+void buf_append_urlencoded(struct oc_text_buf *buf, const char *str)
 {
 	while (str && *str) {
 		unsigned char c = *str;
-		if (c < 0x80 && isalnum((int)(c)))
+		if (c < 0x80 && (isalnum((int)(c)) || c=='-' || c=='_' || c=='.' || c=='~'))
 			buf_append_bytes(buf, str, 1);
 		else
 			buf_append(buf, "%%%02x", c);
@@ -54,6 +54,15 @@ void buf_append_urlencoded(struct oc_text_buf *buf, char *str)
 	}
 }
 
+void buf_append_hex(struct oc_text_buf *buf, const void *str, unsigned len)
+{
+	const unsigned char *data = str;
+	unsigned i;
+
+	for (i = 0; i < len; i++)
+		buf_append(buf, "%02x", (unsigned)data[i]);
+}
+
 void buf_truncate(struct oc_text_buf *buf)
 {
 	if (!buf)
@@ -1295,7 +1304,10 @@ static int process_http_proxy(struct openconnect_info *vpninfo)
  retry:
 	reqbuf = buf_alloc();
 	buf_append(reqbuf, "CONNECT %s:%d HTTP/1.1\r\n", vpninfo->hostname, vpninfo->port);
-	buf_append(reqbuf, "Host: %s\r\n", vpninfo->hostname);
+	if (vpninfo->port == 443)
+		buf_append(reqbuf, "Host: %s\r\n", vpninfo->hostname);
+	else
+		buf_append(reqbuf, "Host: %s:%d\r\n", vpninfo->hostname, vpninfo->port);
 	buf_append(reqbuf, "User-Agent: %s\r\n", vpninfo->useragent);
 	buf_append(reqbuf, "Proxy-Connection: keep-alive\r\n");
 	buf_append(reqbuf, "Connection: keep-alive\r\n");
@@ -1428,7 +1440,10 @@ void http_common_headers(struct openconnect_info *vpninfo, struct oc_text_buf *b
 {
 	struct oc_vpn_option *opt;
 
-	buf_append(buf, "Host: %s\r\n", vpninfo->hostname);
+	if (vpninfo->port == 443)
+		buf_append(buf, "Host: %s\r\n", vpninfo->hostname);
+	else
+		buf_append(buf, "Host: %s:%d\r\n", vpninfo->hostname, vpninfo->port);
 	buf_append(buf, "User-Agent: %s\r\n", vpninfo->useragent);
 
 	if (vpninfo->cookies) {
