@@ -68,6 +68,8 @@ struct openconnect_info *openconnect_vpninfo_new(const char *useragent,
 	init_pkt_queue(&vpninfo->incoming_queue);
 	init_pkt_queue(&vpninfo->outgoing_queue);
 	init_pkt_queue(&vpninfo->oncp_control_queue);
+	vpninfo->dtls_tos_current = 0;
+	vpninfo->dtls_pass_tos = 0;
 	vpninfo->ssl_fd = vpninfo->dtls_fd = -1;
 	vpninfo->cmd_fd = vpninfo->cmd_fd_write = -1;
 	vpninfo->tncc_fd = -1;
@@ -125,7 +127,7 @@ const struct vpn_proto openconnect_protos[] = {
 		.tcp_mainloop = oncp_mainloop,
 		.add_http_headers = oncp_common_headers,
 		.obtain_cookie = oncp_obtain_cookie,
-#if defined(ESP_GNUTLS) || defined(ESP_OPENSSL)
+#ifdef HAVE_ESP
 		.udp_setup = esp_setup,
 		.udp_mainloop = esp_mainloop,
 		.udp_close = esp_close,
@@ -153,8 +155,12 @@ int openconnect_set_protocol(struct openconnect_info *vpninfo, const char *proto
 	return -EINVAL;
 }
 
-void openconnect_set_loglevel(struct openconnect_info *vpninfo,
-			      int level)
+void openconnect_set_pass_tos(struct openconnect_info *vpninfo, int enable)
+{
+	vpninfo->dtls_pass_tos = enable;
+}
+
+void openconnect_set_loglevel(struct openconnect_info *vpninfo, int level)
 {
 	vpninfo->verbose = level;
 }
@@ -284,9 +290,9 @@ void openconnect_vpninfo_free(struct openconnect_info *vpninfo)
 	free(vpninfo->dtls_cipher);
 #ifdef OPENCONNECT_GNUTLS
 	gnutls_free(vpninfo->cstp_cipher); /* In OpenSSL this is const */
-#endif
-#ifdef DTLS_GNUTLS
+#ifdef HAVE_DTLS
 	gnutls_free(vpninfo->gnutls_dtls_cipher);
+#endif
 #endif
 	free(vpninfo->dtls_addr);
 
@@ -333,7 +339,8 @@ void openconnect_vpninfo_free(struct openconnect_info *vpninfo)
 		free(cache);
 	}
 
-	free(vpninfo->peer_cert_hash);
+	free(vpninfo->peer_cert_sha1);
+	free(vpninfo->peer_cert_sha256);
 	free(vpninfo->localname);
 	free(vpninfo->useragent);
 	free(vpninfo->authgroup);
@@ -551,6 +558,7 @@ void openconnect_reset_ssl(struct openconnect_info *vpninfo)
 
 	free(vpninfo->peer_addr);
 	vpninfo->peer_addr = NULL;
+	vpninfo->dtls_tos_optname = 0;
 	free(vpninfo->ip_info.gateway_addr);
 	vpninfo->ip_info.gateway_addr = NULL;
 
@@ -833,6 +841,10 @@ int openconnect_setup_tun_device(struct openconnect_info *vpninfo,
 	if (tun_fd < 0)
 		return tun_fd;
 
+#ifdef _WIN32
+	if (vpninfo->tun_idx != -1)
+		script_setenv_int(vpninfo, "TUNIDX", vpninfo->tun_idx);
+#endif
 	legacy_ifname = openconnect_utf8_to_legacy(vpninfo, vpninfo->ifname);
 	script_setenv(vpninfo, "TUNDEV", legacy_ifname, 0);
 	if (legacy_ifname != vpninfo->ifname)
@@ -866,7 +878,7 @@ const char *openconnect_get_dtls_compression(struct openconnect_info * vpninfo)
 
 const char *openconnect_get_dtls_cipher(struct openconnect_info *vpninfo)
 {
-#if defined(DTLS_GNUTLS)
+#if defined(OPENCONNECT_GNUTLS)
 	if (vpninfo->dtls_state != DTLS_CONNECTED) {
 		gnutls_free(vpninfo->gnutls_dtls_cipher);
 		vpninfo->gnutls_dtls_cipher = NULL;
@@ -878,7 +890,10 @@ const char *openconnect_get_dtls_cipher(struct openconnect_info *vpninfo)
 		vpninfo->gnutls_dtls_cipher = get_gnutls_cipher(vpninfo->dtls_ssl);
 	return vpninfo->gnutls_dtls_cipher;
 #else
-	return vpninfo->dtls_cipher;
+	if (vpninfo->dtls_ssl)
+		return SSL_get_cipher(vpninfo->dtls_ssl);
+	else
+		return NULL;
 #endif
 }
 
@@ -926,9 +941,22 @@ int openconnect_check_peer_cert_hash(struct openconnect_info *vpninfo,
 {
 	char sha1_text[41];
 	const char *fingerprint;
+	unsigned min_match_len;
+	unsigned real_min_match_len = 4;
+	unsigned old_len, fingerprint_len;
 
 	if (strchr(old_hash, ':')) {
-		fingerprint = openconnect_get_peer_cert_hash(vpninfo);
+		if (strncmp(old_hash, "sha1:", 5) == 0) {
+			fingerprint = vpninfo->peer_cert_sha1;
+			min_match_len = real_min_match_len + sizeof("sha1:")-1;
+		} else if (strncmp(old_hash, "sha256:", 7) == 0) {
+			fingerprint = vpninfo->peer_cert_sha256;
+			min_match_len = real_min_match_len + sizeof("sha256:")-1;
+		} else {
+			vpn_progress(vpninfo, PRG_ERR, _("Unknown certificate hash: %s.\n"), old_hash);
+			return -EIO;
+		}
+
 		if (!fingerprint)
 			return -EIO;
 	} else {
@@ -947,10 +975,24 @@ int openconnect_check_peer_cert_hash(struct openconnect_info *vpninfo,
 			sprintf(&sha1_text[i*2], "%02x", sha1_bin[i]);
 
 		fingerprint = sha1_text;
+		min_match_len = real_min_match_len;
 	}
 
-	if (strcasecmp(old_hash, fingerprint))
-		return 1;
+	old_len = strlen(old_hash);
+	fingerprint_len = strlen(fingerprint);
+
+	/* allow partial matches */
+	if (old_len < fingerprint_len) {
+		if (strncasecmp(old_hash, fingerprint, MAX(min_match_len, old_len))) {
+			if (old_len < min_match_len) {
+				vpn_progress(vpninfo, PRG_ERR, _("The size of the provided fingerprint is less than the minimum required (%u).\n"), real_min_match_len);
+			}
+			return 1;
+		}
+	} else {
+		if (strcasecmp(old_hash, fingerprint))
+			return 1;
+	}
 
 	return 0;
 }
@@ -962,7 +1004,7 @@ const char *openconnect_get_cstp_cipher(struct openconnect_info *vpninfo)
 
 const char *openconnect_get_peer_cert_hash(struct openconnect_info *vpninfo)
 {
-	return vpninfo->peer_cert_hash;
+	return vpninfo->peer_cert_sha256;
 }
 
 int openconnect_set_compression_mode(struct openconnect_info *vpninfo,
