@@ -82,7 +82,6 @@ static int verbose = PRG_INFO;
 static int timestamp;
 int background;
 static int do_passphrase_from_fsid;
-static int nocertcheck;
 static int non_inter;
 static int cookieonly;
 static int allow_stdin_read;
@@ -188,6 +187,7 @@ enum {
 	OPT_HTTP_AUTH,
 	OPT_LOCAL_HOSTNAME,
 	OPT_PROTOCOL,
+	OPT_PASSTOS,
 };
 
 #ifdef __sun__
@@ -227,6 +227,7 @@ static const struct option long_options[] = {
 	OPTION("base-mtu", 1, OPT_BASEMTU),
 	OPTION("script", 1, 's'),
 	OPTION("timestamp", 0, OPT_TIMESTAMP),
+	OPTION("passtos", 0, OPT_PASSTOS),
 	OPTION("key-password", 1, 'p'),
 	OPTION("proxy", 1, 'P'),
 	OPTION("proxy-auth", 1, OPT_PROXY_AUTH),
@@ -366,17 +367,29 @@ static char *convert_arg_to_utf8(char **argv, char *arg)
 static void read_stdin(char **string, int hidden, int allow_fail)
 {
 	CONSOLE_READCONSOLE_CONTROL rcc = { sizeof(rcc), 0, 13, 0 };
-	HANDLE stdinh = GetStdHandle(STD_INPUT_HANDLE);
+	HANDLE conh, stdinh = GetStdHandle(STD_INPUT_HANDLE);
 	DWORD cmode, nr_read;
 	wchar_t wbuf[1024];
 	char *buf;
 
+	conh = stdinh;
+	if (!GetConsoleMode(conh, &cmode)) {
+		/* STDIN is not a console? Try opening it explicitly */
+		conh = CreateFile("CONIN$", GENERIC_READ, FILE_SHARE_READ,
+				  0,OPEN_EXISTING, 0, 0);
+		if (conh == INVALID_HANDLE_VALUE || !GetConsoleMode(conh, &cmode)) {
+			char *errstr = openconnect__win32_strerror(GetLastError());
+			fprintf(stderr, _("Failed to open CONIN$: %s\n"), errstr);
+			free(errstr);
+			*string = NULL;
+			goto out;
+		}
+	}
 	if (hidden) {
-		GetConsoleMode(stdinh, &cmode);
-		SetConsoleMode(stdinh, cmode & (~ENABLE_ECHO_INPUT));
+		SetConsoleMode(conh, cmode & (~ENABLE_ECHO_INPUT));
 	}
 
-	if (!ReadConsoleW(stdinh, wbuf, sizeof(wbuf)/2, &nr_read, &rcc)) {
+	if (!ReadConsoleW(conh, wbuf, sizeof(wbuf)/2, &nr_read, &rcc)) {
 		char *errstr = openconnect__win32_strerror(GetLastError());
 		fprintf(stderr, _("ReadConsole() failed: %s\n"), errstr);
 		free(errstr);
@@ -416,9 +429,11 @@ static void read_stdin(char **string, int hidden, int allow_fail)
 
 out:
 	if (hidden) {
-		SetConsoleMode(stdinh, cmode);
+		SetConsoleMode(conh, cmode);
 		fprintf(stderr, "\n");
 	}
+	if (conh != stdinh && conh != INVALID_HANDLE_VALUE)
+		CloseHandle(conh);
 }
 
 #elif defined(HAVE_ICONV)
@@ -616,11 +631,7 @@ static void print_build_opts(void)
 	}
 
 #ifdef HAVE_DTLS
-	printf("%sDTLS", sep);
-#if defined(OPENCONNECT_GNUTLS) && defined(DTLS_OPENSSL)
-	printf(" (%s)", _("using OpenSSL"));
-#endif
-	printf("\n");
+	printf("%sDTLS\n", sep);
 #else
 	printf(_("\nWARNING: No DTLS support in this binary. Performance will be impaired.\n"));
 #endif
@@ -761,12 +772,13 @@ static void usage(void)
 	printf("  -l, --syslog                    %s\n", _("Use syslog for progress messages"));
 #endif
 	printf("      --timestamp                 %s\n", _("Prepend timestamp to progress messages"));
+	printf("      --passtos                   %s\n", _("copy TOS / TCLASS when using DTLS"));
 #ifndef _WIN32
 	printf("  -U, --setuid=USER               %s\n", _("Drop privileges after connecting"));
 	printf("      --csd-user=USER             %s\n", _("Drop privileges during CSD execution"));
 	printf("      --csd-wrapper=SCRIPT        %s\n", _("Run SCRIPT instead of CSD binary"));
 #endif
-	printf("  -m, --mtu=MTU                   %s\n", _("Request MTU from server"));
+	printf("  -m, --mtu=MTU                   %s\n", _("Request MTU from server (legacy servers only)"));
 	printf("      --base-mtu=MTU              %s\n", _("Indicate path MTU to/from server"));
 	printf("  -p, --key-password=PASS         %s\n", _("Set key passphrase or TPM SRK PIN"));
 	printf("      --key-password-from-fsid    %s\n", _("Key passphrase is fsid of file system"));
@@ -1288,7 +1300,10 @@ int main(int argc, char **argv)
 			vpninfo->no_http_keepalive = 1;
 			break;
 		case OPT_NO_CERT_CHECK:
-			nocertcheck = 1;
+			fprintf(stderr,
+				_("The --no-cert-check option was insecure and has been removed.\n"
+				  "Fix your server's certificate or use --servercert to trust it.\n"));
+			exit(1);
 			break;
 		case 's':
 			vpnc_script = dup_config_arg();
@@ -1372,6 +1387,9 @@ int main(int argc, char **argv)
 					xstrdup("AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA"));
 			}
 			break;
+		case OPT_PASSTOS:
+			openconnect_set_pass_tos(vpninfo, 1);
+			break;
 		case OPT_TIMESTAMP:
 			timestamp = 1;
 			break;
@@ -1693,9 +1711,6 @@ static int validate_peer_cert(void *_vpninfo, const char *reason)
 		return -EINVAL;
 	}
 
-	if (nocertcheck)
-		return 0;
-
 	fingerprint = openconnect_get_peer_cert_hash(vpninfo);
 
 	for (this = accepted_certs; this; this = this->next) {
@@ -1712,6 +1727,9 @@ static int validate_peer_cert(void *_vpninfo, const char *reason)
 		fprintf(stderr, _("\nCertificate from VPN server \"%s\" failed verification.\n"
 			 "Reason: %s\n"), vpninfo->hostname, reason);
 
+		fprintf(stderr, _("To trust this server in future, perhaps add this to your command line:\n"));
+		fprintf(stderr, _("    --servercert %s\n"), fingerprint);
+
 		if (non_inter)
 			return -EINVAL;
 
