@@ -433,6 +433,7 @@ static const unsigned char data_hdr[] = { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
 					  0x01, 0x2c, 0x01, 0x00, 0x00, 0x00,
 					  0x01, 0x00, 0x00, 0x00, 0x00, 0x00 };
 
+#ifdef HAVE_ESP
 static const unsigned char esp_kmp_hdr[] = {
 	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x2e,
 	0x01, 0x00, 0x00, 0x00, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, /* KMP header */
@@ -445,14 +446,14 @@ static const unsigned char esp_kmp_part2[] = {
 	0x00, 0x02, 0x00, 0x00, 0x00, 0x40, /* Attr 2 (secrets) */
 };
 /* And now 0x40 bytes of random secret for encryption and HMAC key */
-
+#endif
 
 static const struct pkt esp_enable_pkt = {
 	.next = NULL,
-	{ .oncp.rec = { 0x21, 0x00 },
-	  .oncp.kmp = { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x2f,
-			0x01, 0x00, 0x00, 0x00, 0x01, 0x00, 0x00, 0x00,
-			0x00, 0x00, 0x00, 0x0d }
+	{ .oncp = { .rec = { 0x21, 0x00 },
+	            .kmp = { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x2f,
+			     0x01, 0x00, 0x00, 0x00, 0x01, 0x00, 0x00, 0x00,
+			     0x00, 0x00, 0x00, 0x0d } }
 	},
 	.data = {
 		0x00, 0x06, 0x00, 0x00, 0x00, 0x07, /* Group 6, len 7 */
@@ -542,7 +543,7 @@ int oncp_connect(struct openconnect_info *vpninfo)
 {
 	int ret, len, kmp, kmplen, group;
 	struct oc_text_buf *reqbuf;
-	unsigned char bytes[8192];
+	unsigned char bytes[16384];
 
 	/* XXX: We should do what cstp_connect() does to check that configuration
 	   hasn't changed on a reconnect. */
@@ -561,6 +562,7 @@ int oncp_connect(struct openconnect_info *vpninfo)
 
  	buf_append(reqbuf, "POST /dana/js?prot=1&svc=1 HTTP/1.1\r\n");
 	oncp_common_headers(vpninfo, reqbuf);
+	buf_append(reqbuf, "Content-Length: 256\r\n");
 	buf_append(reqbuf, "\r\n");
 
 	if (buf_error(reqbuf)) {
@@ -605,6 +607,7 @@ int oncp_connect(struct openconnect_info *vpninfo)
 	buf_truncate(reqbuf);
 	buf_append(reqbuf, "POST /dana/js?prot=1&svc=4 HTTP/1.1\r\n");
 	oncp_common_headers(vpninfo, reqbuf);
+	buf_append(reqbuf, "Content-Length: 256\r\n");
 	buf_append(reqbuf, "\r\n");
 
 	if (buf_error(reqbuf)) {
@@ -774,7 +777,7 @@ int oncp_connect(struct openconnect_info *vpninfo)
 	put_len32(reqbuf, group);
 	put_len16(reqbuf, kmp);
 
-#if defined(ESP_GNUTLS) || defined(ESP_OPENSSL)
+#ifdef HAVE_ESP
 	if (!setup_esp_keys(vpninfo)) {
 		struct esp *esp = &vpninfo->esp_in[vpninfo->current_esp_in];
 		/* Since we'll want to do this in the oncp_mainloop too, where it's easier
@@ -823,7 +826,7 @@ int oncp_connect(struct openconnect_info *vpninfo)
 
 static int oncp_receive_espkeys(struct openconnect_info *vpninfo, int len)
 {
-#if defined(ESP_GNUTLS) || defined(ESP_OPENSSL)
+#ifdef HAVE_ESP
 	int ret;
 
 	ret = parse_conf_pkt(vpninfo, vpninfo->cstp_pkt->oncp.kmp, len + 20, 301);
@@ -999,7 +1002,7 @@ int oncp_mainloop(struct openconnect_info *vpninfo, int *timeout)
 				iplen = load_be16(vpninfo->cstp_pkt->data + 2);
 				break;
 			case 6:
-				iplen = load_be16(vpninfo->cstp_pkt->data + 4);
+				iplen = load_be16(vpninfo->cstp_pkt->data + 4) + 40;
 				break;
 			default:
 			badiplen:
@@ -1098,7 +1101,7 @@ int oncp_mainloop(struct openconnect_info *vpninfo, int *timeout)
 			/* XXX: Do we have to do this or can we leave it open?
 			 * Perhaps we could even reconnect asynchronously while
 			 * the ESP is still running? */
-#if defined(ESP_GNUTLS) || defined(ESP_OPENSSL)
+#ifdef HAVE_ESP
 			esp_shutdown(vpninfo);
 #endif
 			ret = ssl_reconnect(vpninfo);
