@@ -41,7 +41,7 @@
 
 #include "openconnect.h"
 
-#if defined(OPENCONNECT_OPENSSL) || defined(DTLS_OPENSSL)
+#if defined(OPENCONNECT_OPENSSL)
 #include <openssl/ssl.h>
 #include <openssl/err.h>
 /* Ick */
@@ -73,6 +73,7 @@
 #include <sys/time.h>
 #include <sys/types.h>
 #include <unistd.h>
+#include <string.h>
 
 #ifdef LIBPROXY_HDR
 #include LIBPROXY_HDR
@@ -113,10 +114,14 @@
 
 #include <libxml/tree.h>
 
+#define SHA256_SIZE 32
 #define SHA1_SIZE 20
 #define MD5_SIZE 16
 
+/* FreeBSD provides this in <sys/param.h>  */
+#ifndef MAX
 #define MAX(x,y) ((x)>(y))?(x):(y)
+#endif
 /****************************************************************************/
 
 struct pkt {
@@ -171,6 +176,8 @@ struct pkt {
 #endif
 #define COMPR_ALL	(COMPR_STATELESS | COMPR_DEFLATE)
 
+#define DTLS_APP_ID_EXT 48018
+
 struct keepalive_info {
 	int dpd;
 	int keepalive;
@@ -312,17 +319,21 @@ static inline void init_pkt_queue(struct pkt_q *q)
 {
 	q->tail = &q->head;
 }
-	
+
+#define DTLS_OVERHEAD (1 /* packet + header */ + 13 /* DTLS header */ + \
+	 20 /* biggest supported MAC (SHA1) */ +  16 /* biggest supported IV (AES-128) */ + \
+	 16 /* max padding */)
+
 struct esp {
-#if defined(ESP_GNUTLS)
+#if defined(OPENCONNECT_GNUTLS)
 	gnutls_cipher_hd_t cipher;
 	gnutls_hmac_hd_t hmac;
-#elif defined(ESP_OPENSSL)
+#elif defined(OPENCONNECT_OPENSSL)
 	HMAC_CTX *hmac, *pkt_hmac;
 	EVP_CIPHER_CTX *cipher;
 #endif
-	uint32_t seq;
-	uint32_t seq_backlog;
+	uint64_t seq_backlog;
+	uint64_t seq;
 	uint32_t spi; /* Stored network-endian */
 	unsigned char secrets[0x40];
 };
@@ -444,7 +455,8 @@ struct openconnect_info {
 	void *tok_cbdata;
 
 	void *peer_cert;
-	char *peer_cert_hash;
+	char *peer_cert_sha1;
+	char *peer_cert_sha256;
 	void *cert_list_handle;
 	int cert_list_size;
 
@@ -472,6 +484,7 @@ struct openconnect_info {
 #elif defined(OPENCONNECT_GNUTLS)
 	gnutls_session_t https_sess;
 	gnutls_certificate_credentials_t https_cred;
+	gnutls_psk_client_credentials_t psk_cred;
 	char local_cert_md5[MD5_SIZE * 2 + 1]; /* For CSD */
 	char gnutls_prio[256];
 #ifdef HAVE_TROUSERS
@@ -517,10 +530,10 @@ struct openconnect_info {
 	int reconnect_interval;
 	int dtls_attempt_period;
 	time_t new_dtls_started;
-#if defined(DTLS_OPENSSL)
+#if defined(OPENCONNECT_OPENSSL)
 	SSL_CTX *dtls_ctx;
 	SSL *dtls_ssl;
-#elif defined(DTLS_GNUTLS)
+#elif defined(OPENCONNECT_GNUTLS)
 	/* Call this dtls_ssl rather than dtls_sess because it's just a
 	   pointer, and generic code in dtls.c wants to check if it's
 	   NULL or not or pass it to DTLS_SEND/DTLS_RECV. This way we
@@ -535,6 +548,8 @@ struct openconnect_info {
 	struct keepalive_info dtls_times;
 	unsigned char dtls_session_id[32];
 	unsigned char dtls_secret[48];
+	unsigned char dtls_app_id[32];
+	unsigned dtls_app_id_size;
 
 	char *dtls_cipher;
 	char *vpnc_script;
@@ -550,10 +565,11 @@ struct openconnect_info {
 	char *ifname;
 	char *cmd_ifname;
 
-	int reqmtu, basemtu;
+	int reqmtu, basemtu; /* Local static configured values */
 	const char *banner;
 
 	struct oc_ip_info ip_info;
+	int cstp_basemtu; /* Returned by server */
 
 #ifdef _WIN32
 	long dtls_monitored, ssl_monitored, cmd_monitored, tun_monitored;
@@ -572,13 +588,17 @@ struct openconnect_info {
 #ifdef _WIN32
 	HANDLE tun_fh;
 	OVERLAPPED tun_rd_overlap, tun_wr_overlap;
-	int tun_rd_pending;
+	int tun_idx, tun_rd_pending;
 #else
 	int tun_fd;
 #endif
 	int ssl_fd;
 	int dtls_fd;
 
+	int dtls_tos_current;
+	int dtls_pass_tos;
+	int dtls_tos_proto, dtls_tos_optname;
+
 	int cmd_fd;
 	int cmd_fd_write;
 	int got_cancel_cmd;
@@ -649,10 +669,10 @@ struct openconnect_info {
 #define read_fd_monitored(_v, _n) FD_ISSET(_v->_n##_fd, &_v->_select_rfds)
 #endif
 
-#if (defined(DTLS_OPENSSL) && defined(SSL_OP_CISCO_ANYCONNECT)) || \
-    (defined(DTLS_GNUTLS) && defined(HAVE_GNUTLS_SESSION_SET_PREMASTER))
-#define HAVE_DTLS 1
-#endif
+/* Key material for DTLS-PSK */
+#define PSK_LABEL "EXPORTER-openconnect-psk"
+#define PSK_LABEL_SIZE sizeof(PSK_LABEL)-1
+#define PSK_KEY_SIZE 32
 
 /* Packet types */
 
@@ -668,7 +688,7 @@ struct openconnect_info {
 	if ((_v)->verbose >= (lvl))					\
 		(_v)->progress((_v)->cbdata, lvl, __VA_ARGS__);	\
 	} while(0)
-#define vpn_perror(vpninfo, msg) vpn_progress((vpninfo), PRG_ERR, "%s: %s\n", (msg), strerror(errno));
+#define vpn_perror(vpninfo, msg) vpn_progress((vpninfo), PRG_ERR, "%s: %s\n", (msg), strerror(errno))
 
 /****************************************************************************/
 /* Oh Solaris how we hate thee! */
@@ -786,12 +806,22 @@ int os_read_tun(struct openconnect_info *vpninfo, struct pkt *pkt);
 int os_write_tun(struct openconnect_info *vpninfo, struct pkt *pkt);
 intptr_t os_setup_tun(struct openconnect_info *vpninfo);
 
+/* {gnutls,openssl}-dtls.c */
+int start_dtls_handshake(struct openconnect_info *vpninfo, int dtls_fd);
+int dtls_try_handshake(struct openconnect_info *vpninfo);
+unsigned dtls_set_mtu(struct openconnect_info *vpninfo, unsigned mtu);
+void dtls_ssl_free(struct openconnect_info *vpninfo);
+
 /* dtls.c */
 int dtls_setup(struct openconnect_info *vpninfo, int dtls_attempt_period);
 int dtls_mainloop(struct openconnect_info *vpninfo, int *timeout);
 void dtls_close(struct openconnect_info *vpninfo);
 void dtls_shutdown(struct openconnect_info *vpninfo);
 void append_dtls_ciphers(struct openconnect_info *vpninfo, struct oc_text_buf *buf);
+void dtls_detect_mtu(struct openconnect_info *vpninfo);
+int openconnect_dtls_read(struct openconnect_info *vpninfo, void *buf, size_t len, unsigned ms);
+int openconnect_dtls_write(struct openconnect_info *vpninfo, void *buf, size_t len);
+char *openconnect_bin2hex(const char *prefix, const uint8_t *data, unsigned len);
 
 /* cstp.c */
 void cstp_common_headers(struct openconnect_info *vpninfo, struct oc_text_buf *buf);
@@ -873,6 +903,7 @@ int cstp_handshake(struct openconnect_info *vpninfo, unsigned init);
 int get_cert_md5_fingerprint(struct openconnect_info *vpninfo, void *cert,
 			     char *buf);
 int openconnect_sha1(unsigned char *result, void *data, int len);
+int openconnect_sha256(unsigned char *result, void *data, int len);
 int openconnect_md5(unsigned char *result, void *data, int len);
 int openconnect_random(void *bytes, int len);
 int openconnect_local_cert_md5(struct openconnect_info *vpninfo,
@@ -942,7 +973,7 @@ int cstp_obtain_cookie(struct openconnect_info *vpninfo);
 int xmlnode_is_named(xmlNode *xml_node, const char *name);
 int xmlnode_get_prop(xmlNode *xml_node, const char *name, char **var);
 int xmlnode_match_prop(xmlNode *xml_node, const char *name, const char *match);
-int append_opt(struct oc_text_buf *body, char *opt, char *name);
+int append_opt(struct oc_text_buf *body, const char *opt, const char *name);
 int append_form_opts(struct openconnect_info *vpninfo,
 		     struct oc_auth_form *form, struct oc_text_buf *body);
 void free_opt(struct oc_form_opt *opt);
@@ -960,11 +991,12 @@ int buf_ensure_space(struct oc_text_buf *buf, int len);
 void  __attribute__ ((format (printf, 2, 3)))
 	buf_append(struct oc_text_buf *buf, const char *fmt, ...);
 void buf_append_bytes(struct oc_text_buf *buf, const void *bytes, int len);
+void buf_append_hex(struct oc_text_buf *buf, const void *str, unsigned len);
 int buf_append_utf16le(struct oc_text_buf *buf, const char *utf8);
 int get_utf8char(const char **utf8);
 void buf_append_from_utf16le(struct oc_text_buf *buf, const void *utf16);
 void buf_truncate(struct oc_text_buf *buf);
-void buf_append_urlencoded(struct oc_text_buf *buf, char *str);
+void buf_append_urlencoded(struct oc_text_buf *buf, const char *str);
 int buf_error(struct oc_text_buf *buf);
 int buf_free(struct oc_text_buf *buf);
 char *openconnect_create_useragent(const char *base);
