@@ -23,6 +23,7 @@
 #include <ctype.h>
 
 #include "openconnect-internal.h"
+#include <openssl/rand.h>
 
 #ifdef HAVE_LIBP11 /* And p11-kit */
 
@@ -101,7 +102,7 @@ static int parse_uri_attr(const char *attr, int attrlen, unsigned char **field,
 
 static int parse_pkcs11_uri(const char *uri, PKCS11_TOKEN **p_tok,
 			    unsigned char **id, size_t *id_len,
-			    char **label)
+			    char **label, char **pin)
 {
 	PKCS11_TOKEN *tok;
 	char *newlabel = NULL;
@@ -149,6 +150,16 @@ static int parse_pkcs11_uri(const char *uri, PKCS11_TOKEN **p_tok,
 			} else
 				ret = -EINVAL;
 			/* Ignore object type for now. */
+		} else if (!strncmp(p, "pin-value=", 10)) {
+			/* XXX We could do better than this but it'll cover all sane
+			   use cases. */
+			char *pinvalue = NULL;
+			p += 10;
+			ret = parse_uri_attr(p, end - p, (void *)&pinvalue, NULL);
+			if (pinvalue) {
+				free(*pin);
+				*pin = pinvalue;
+			}
 		} else {
 			ret = -EINVAL;
 		}
@@ -176,6 +187,11 @@ static int request_pin(struct openconnect_info *vpninfo, struct pin_cache *cache
 	if (!vpninfo || !vpninfo->process_auth_form)
 		return -EINVAL;
 
+	if (vpninfo->cert_password) {
+		cache->pin = vpninfo->cert_password;
+		vpninfo->cert_password = NULL;
+		return 0;
+	}
 	memset(&f, 0, sizeof(f));
 	f.auth_id = (char *)"pkcs11_pin";
 	f.opts = &o;
@@ -308,7 +324,7 @@ int load_pkcs11_certificate(struct openconnect_info *vpninfo)
 {
 	PKCS11_CTX *ctx;
 	PKCS11_TOKEN *match_tok = NULL;
-	PKCS11_CERT *cert;
+	PKCS11_CERT *cert = NULL;
 	char *cert_label = NULL;
 	unsigned char *cert_id = NULL;
 	size_t cert_id_len = 0;
@@ -321,7 +337,7 @@ int load_pkcs11_certificate(struct openconnect_info *vpninfo)
 		return -EIO;
 
 	if (parse_pkcs11_uri(vpninfo->cert, &match_tok, &cert_id,
-			     &cert_id_len, &cert_label) < 0) {
+			     &cert_id_len, &cert_label, &vpninfo->cert_password) < 0) {
 		vpn_progress(vpninfo, PRG_ERR,
 			     _("Failed to parse PKCS#11 URI '%s'\n"),
 			     vpninfo->cert);
@@ -373,6 +389,9 @@ int load_pkcs11_certificate(struct openconnect_info *vpninfo)
 		}
 	}
 	ret = -EINVAL;
+	vpn_progress(vpninfo, PRG_ERR,
+		     _("Failed to find PKCS#11 cert '%s'\n"),
+		     vpninfo->cert);
  got_cert:
 	if (cert) {
 		/* This happens if the cert is too large for the fixed buffer
@@ -459,6 +478,70 @@ static PKCS11_KEY *slot_find_key(struct openconnect_info *vpninfo, PKCS11_CTX *c
 	return NULL;
 }
 
+#ifndef OPENSSL_NO_EC
+#if OPENSSL_VERSION_NUMBER < 0x10100000L || defined(LIBRESSL_VERSION_NUMBER)
+#define EVP_PKEY_id(k) ((k)->type)
+#endif
+static int validate_ecdsa_key(struct openconnect_info *vpninfo, EC_KEY *priv_ec)
+{
+	EVP_PKEY *pub_pkey;
+	EC_KEY *pub_ec;
+	unsigned char rdata[SHA1_SIZE];
+	unsigned int siglen = ECDSA_size(priv_ec);
+	unsigned char *sig;
+	int ret = -EINVAL;
+
+	pub_pkey = X509_get_pubkey(vpninfo->cert_x509);
+	if (!pub_pkey) {
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Certificate has no public key\n"));
+		goto out;
+	}
+	pub_ec = EVP_PKEY_get1_EC_KEY(pub_pkey);
+	if (!pub_ec) {
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Certificate does not match private key\n"));
+		goto out_pkey;
+	}
+	vpn_progress(vpninfo, PRG_TRACE, _("Checking EC key matches cert\n"));
+	sig = malloc(siglen);
+	if (!sig) {
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Failed to allocate signature buffer\n"));
+		ret = -ENOMEM;
+		goto out_pubec;
+	}
+	if (!RAND_bytes(rdata, sizeof(rdata))) {
+		/* Actually, who cares? */
+	}
+	if (!ECDSA_sign(NID_sha1, rdata, sizeof(rdata),
+			sig, &siglen, priv_ec)) {
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Failed to sign dummy data to validate EC key\n"));
+		openconnect_report_ssl_errors(vpninfo);
+		goto out_sig;
+	}
+	if (!ECDSA_verify(NID_sha1, rdata, sizeof(rdata), sig, siglen, pub_ec)) {
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Certificate does not match private key\n"));
+		goto out_sig;
+	}
+
+	/* Finally, copy the public EC_POINT data now that we know it really did match */
+	EC_KEY_set_public_key(priv_ec, EC_KEY_get0_public_key(pub_ec));
+	ret = 0;
+
+ out_sig:
+	free(sig);
+ out_pubec:
+	EC_KEY_free(pub_ec);
+ out_pkey:
+	EVP_PKEY_free(pub_pkey);
+ out:
+	return ret;
+}
+#endif
+
 int load_pkcs11_key(struct openconnect_info *vpninfo)
 {
 	PKCS11_CTX *ctx;
@@ -477,7 +560,7 @@ int load_pkcs11_key(struct openconnect_info *vpninfo)
 		return -EIO;
 
 	if (parse_pkcs11_uri(vpninfo->sslkey, &match_tok, &key_id,
-			     &key_id_len, &key_label) < 0) {
+			     &key_id_len, &key_label, &vpninfo->cert_password) < 0) {
 		vpn_progress(vpninfo, PRG_ERR,
 			     _("Failed to parse PKCS#11 URI '%s'\n"),
 			     vpninfo->sslkey);
@@ -551,11 +634,14 @@ int load_pkcs11_key(struct openconnect_info *vpninfo)
 		}
 	}
 	ret = -EINVAL;
+	vpn_progress(vpninfo, PRG_ERR,
+		     _("Failed to find PKCS#11 key '%s'\n"),
+		     vpninfo->sslkey);
 
  got_key:
 	if (key) {
 		vpn_progress(vpninfo, PRG_DEBUG,
-			     _("Using PKCS#11 key %s\n"), vpninfo->cert);
+			     _("Using PKCS#11 key %s\n"), vpninfo->sslkey);
 
 		pkey = PKCS11_get_private_key(key);
 		if (!pkey) {
@@ -566,6 +652,28 @@ int load_pkcs11_key(struct openconnect_info *vpninfo)
 			goto out;
 		}
 
+#ifndef OPENSSL_NO_EC
+		/*
+		 * If an EC EVP_PKEY has no public key, OpenSSL will crash
+		 * when trying to check it matches the certificate:
+		 * https://github.com/openssl/openssl/issues/1532
+		 *
+		 * Work around this by detecting this condition, manually
+		 * checking that the certificate *does* match by performing
+		 * a signature and validating it against the cert, then
+		 * copying the EC_POINT public key information from the cert.
+		 */
+		if (EVP_PKEY_id(pkey) == EVP_PKEY_EC) {
+			EC_KEY *priv_ec = EVP_PKEY_get1_EC_KEY(pkey);
+
+			ret = 0;
+			if (!EC_KEY_get0_public_key(priv_ec))
+				ret = validate_ecdsa_key(vpninfo, priv_ec);
+			EC_KEY_free(priv_ec);
+			if (ret)
+				goto out;
+		}
+#endif
 		if (!SSL_CTX_use_PrivateKey(vpninfo->https_ctx, pkey)) {
 			vpn_progress(vpninfo, PRG_ERR, _("Add key from PKCS#11 failed\n"));
 			openconnect_report_ssl_errors(vpninfo);
