@@ -36,16 +36,17 @@
 #include <openssl/ui.h>
 #include <openssl/rsa.h>
 
-#if OPENSSL_VERSION_NUMBER < 0x10100000L
+#if OPENSSL_VERSION_NUMBER < 0x10100000L || defined(LIBRESSL_VERSION_NUMBER)
 #define X509_up_ref(x) 	CRYPTO_add(&(x)->references, 1, CRYPTO_LOCK_X509)
-#endif
-
-#if OPENSSL_VERSION_NUMBER < 0x10100000L
+#define X509_get0_notAfter(x) X509_get_notAfter(x)
 #define EVP_MD_CTX_new EVP_MD_CTX_create
 #define EVP_MD_CTX_free EVP_MD_CTX_destroy
 #define X509_STORE_CTX_get0_chain(ctx) ((ctx)->chain)
 #define X509_STORE_CTX_get0_untrusted(ctx) ((ctx)->untrusted)
 #define X509_STORE_CTX_get0_cert(ctx) ((ctx)->cert)
+typedef int (*X509_STORE_CTX_get_issuer_fn)(X509 **issuer,
+					    X509_STORE_CTX *ctx, X509 *x);
+#define X509_STORE_CTX_get_get_issuer(ctx) ((ctx)->get_issuer)
 #endif
 
 int openconnect_sha1(unsigned char *result, void *data, int len)
@@ -61,6 +62,19 @@ int openconnect_sha1(unsigned char *result, void *data, int len)
 	return 0;
 }
 
+int openconnect_sha256(unsigned char *result, void *data, int len)
+{
+	EVP_MD_CTX *c = EVP_MD_CTX_new();
+
+	if (!c)
+		return -ENOMEM;
+
+	EVP_Digest(data, len, result, NULL, EVP_sha256(), NULL);
+	EVP_MD_CTX_free(c);
+
+	return 0;
+}
+
 int openconnect_md5(unsigned char *result, void *data, int len)
 {
 	EVP_MD_CTX *c = EVP_MD_CTX_new();
@@ -108,27 +122,27 @@ int openconnect_random(void *bytes, int len)
 /* Helper functions for reading/writing lines over SSL.
    We could use cURL for the HTTP stuff, but it's overkill */
 
-static int openconnect_openssl_write(struct openconnect_info *vpninfo, char *buf, size_t len)
+static int _openconnect_openssl_write(SSL *ssl, int fd, struct openconnect_info *vpninfo, char *buf, size_t len)
 {
 	size_t orig_len = len;
 
 	while (len) {
-		int done = SSL_write(vpninfo->https_ssl, buf, len);
+		int done = SSL_write(ssl, buf, len);
 
 		if (done > 0)
 			len -= done;
 		else {
-			int err = SSL_get_error(vpninfo->https_ssl, done);
+			int err = SSL_get_error(ssl, done);
 			fd_set wr_set, rd_set;
-			int maxfd = vpninfo->ssl_fd;
+			int maxfd = fd;
 
 			FD_ZERO(&wr_set);
 			FD_ZERO(&rd_set);
 
 			if (err == SSL_ERROR_WANT_READ)
-				FD_SET(vpninfo->ssl_fd, &rd_set);
+				FD_SET(fd, &rd_set);
 			else if (err == SSL_ERROR_WANT_WRITE)
-				FD_SET(vpninfo->ssl_fd, &wr_set);
+				FD_SET(fd, &wr_set);
 			else {
 				vpn_progress(vpninfo, PRG_ERR, _("Failed to write to SSL socket\n"));
 				openconnect_report_ssl_errors(vpninfo);
@@ -145,37 +159,69 @@ static int openconnect_openssl_write(struct openconnect_info *vpninfo, char *buf
 	return orig_len;
 }
 
-static int openconnect_openssl_read(struct openconnect_info *vpninfo, char *buf, size_t len)
+static int openconnect_openssl_write(struct openconnect_info *vpninfo, char *buf, size_t len)
 {
-	int done;
+	return _openconnect_openssl_write(vpninfo->https_ssl, vpninfo->ssl_fd, vpninfo, buf, len);
+}
 
-	while ((done = SSL_read(vpninfo->https_ssl, buf, len)) == -1) {
-		int err = SSL_get_error(vpninfo->https_ssl, done);
+int openconnect_dtls_write(struct openconnect_info *vpninfo, void *buf, size_t len)
+{
+	return _openconnect_openssl_write(vpninfo->dtls_ssl, vpninfo->dtls_fd, vpninfo, buf, len);
+}
+
+/* set ms to zero for no timeout */
+static int _openconnect_openssl_read(SSL *ssl, int fd, struct openconnect_info *vpninfo, char *buf, size_t len, unsigned ms)
+{
+	int done, ret;
+	struct timeval timeout, *tv = NULL;
+
+	if (ms) {
+		timeout.tv_sec = ms/1000;
+		timeout.tv_usec = (ms%1000)*1000;
+		tv = &timeout;
+	}
+
+	while ((done = SSL_read(ssl, buf, len)) == -1) {
+		int err = SSL_get_error(ssl, done);
 		fd_set wr_set, rd_set;
-		int maxfd = vpninfo->ssl_fd;
+		int maxfd = fd;
 
 		FD_ZERO(&wr_set);
 		FD_ZERO(&rd_set);
 
 		if (err == SSL_ERROR_WANT_READ)
-			FD_SET(vpninfo->ssl_fd, &rd_set);
+			FD_SET(fd, &rd_set);
 		else if (err == SSL_ERROR_WANT_WRITE)
-			FD_SET(vpninfo->ssl_fd, &wr_set);
+			FD_SET(fd, &wr_set);
 		else {
 			vpn_progress(vpninfo, PRG_ERR, _("Failed to read from SSL socket\n"));
 			openconnect_report_ssl_errors(vpninfo);
 			return -EIO;
 		}
 		cmd_fd_set(vpninfo, &rd_set, &maxfd);
-		select(maxfd + 1, &rd_set, &wr_set, NULL, NULL);
+		ret = select(maxfd + 1, &rd_set, &wr_set, NULL, tv);
 		if (is_cancel_pending(vpninfo, &rd_set)) {
 			vpn_progress(vpninfo, PRG_ERR, _("SSL read cancelled\n"));
 			return -EINTR;
 		}
+
+		if (ret == 0) {
+			return -ETIMEDOUT;
+		}
 	}
 	return done;
 }
 
+static int openconnect_openssl_read(struct openconnect_info *vpninfo, char *buf, size_t len)
+{
+	return _openconnect_openssl_read(vpninfo->https_ssl, vpninfo->ssl_fd, vpninfo, buf, len, 0);
+}
+
+int openconnect_dtls_read(struct openconnect_info *vpninfo, void *buf, size_t len, unsigned ms)
+{
+	return _openconnect_openssl_read(vpninfo->dtls_ssl, vpninfo->dtls_fd, vpninfo, buf, len, ms);
+}
+
 static int openconnect_openssl_gets(struct openconnect_info *vpninfo, char *buf, size_t len)
 {
 	int i = 0;
@@ -467,6 +513,8 @@ static int install_extra_certs(struct openconnect_info *vpninfo, const char *sou
 	X509 *cert = vpninfo->cert_x509;
 	int i;
 
+	if (!cert)
+		return -EINVAL;
  next:
 	for (i = 0; i < sk_X509_num(ca); i++) {
 		X509 *cert2 = sk_X509_value(ca, i);
@@ -546,7 +594,7 @@ static int load_pkcs12_certificate(struct openconnect_info *vpninfo, PKCS12 *p12
 			     _("Using client certificate '%s'\n"), buf);
 	} else {
 		vpn_progress(vpninfo, PRG_ERR,
-			     _("PKCS#12 contained no certificate!"));
+			     _("PKCS#12 contained no certificate!\n"));
 		ret = -EINVAL;
 	}
 
@@ -555,7 +603,7 @@ static int load_pkcs12_certificate(struct openconnect_info *vpninfo, PKCS12 *p12
 		EVP_PKEY_free(pkey);
 	} else {
 		vpn_progress(vpninfo, PRG_ERR,
-			     _("PKCS#12 contained no private key!"));
+			     _("PKCS#12 contained no private key!\n"));
 		ret = -EINVAL;
 	}
 
@@ -791,8 +839,10 @@ static int is_pem_password_error(struct openconnect_info *vpninfo)
 
 static int load_certificate(struct openconnect_info *vpninfo)
 {
+	EVP_PKEY *key;
 	FILE *f;
 	char buf[256];
+	int ret;
 
 	if (!strncmp(vpninfo->cert, "pkcs11:", 7)) {
 		int ret = load_pkcs11_certificate(vpninfo);
@@ -856,7 +906,6 @@ static int load_certificate(struct openconnect_info *vpninfo)
  got_cert:
 #ifdef ANDROID_KEYSTORE
 	if (!strncmp(vpninfo->sslkey, "keystore:", 9)) {
-		EVP_PKEY *key;
 		BIO *b;
 
 	again_android:
@@ -878,6 +927,7 @@ static int load_certificate(struct openconnect_info *vpninfo)
 			vpninfo->cert_x509 = NULL;
 			return -EINVAL;
 		}
+		EVP_PKEY_free(key);
 		return 0;
 	}
 #endif /* ANDROID_KEYSTORE */
@@ -888,7 +938,7 @@ static int load_certificate(struct openconnect_info *vpninfo)
 	if (!f) {
 		vpn_progress(vpninfo, PRG_ERR,
 			     _("Failed to open private key file %s: %s\n"),
-			     vpninfo->cert, strerror(errno));
+			     vpninfo->sslkey, strerror(errno));
 		return -ENOENT;
 	}
 
@@ -899,9 +949,9 @@ static int load_certificate(struct openconnect_info *vpninfo)
 			return load_tpm_certificate(vpninfo);
 		} else if (!strcmp(buf, "-----BEGIN RSA PRIVATE KEY-----\n") ||
 			   !strcmp(buf, "-----BEGIN DSA PRIVATE KEY-----\n") ||
+			   !strcmp(buf, "-----BEGIN EC PRIVATE KEY-----\n") ||
 			   !strcmp(buf, "-----BEGIN ENCRYPTED PRIVATE KEY-----\n") ||
 			   !strcmp(buf, "-----BEGIN PRIVATE KEY-----\n")) {
-			RSA *key;
 			BIO *b = BIO_new_fp(f, BIO_CLOSE);
 
 			if (!b) {
@@ -912,21 +962,113 @@ static int load_certificate(struct openconnect_info *vpninfo)
 			}
 		again:
 			fseek(f, 0, SEEK_SET);
-			key = PEM_read_bio_RSAPrivateKey(b, NULL, pem_pw_cb, vpninfo);
+			key = PEM_read_bio_PrivateKey(b, NULL, pem_pw_cb, vpninfo);
 			if (!key) {
 				if (is_pem_password_error(vpninfo))
 					goto again;
 				BIO_free(b);
 				return -EINVAL;
 			}
-			SSL_CTX_use_RSAPrivateKey(vpninfo->https_ctx, key);
-			RSA_free(key);
+			ret = 0;
+			if (!SSL_CTX_use_PrivateKey(vpninfo->https_ctx, key)) {
+				vpn_progress(vpninfo, PRG_ERR,
+					     _("Loading private key failed\n"));
+				openconnect_report_ssl_errors(vpninfo);
+				ret = -EINVAL;
+			}
+			EVP_PKEY_free(key);
 			BIO_free(b);
-			return 0;
+			return ret;
 		}
 	}
-	fclose(f);
 
+	/* Not PEM? Try DER... */
+	fseek(f, 0, SEEK_SET);
+
+	/* This will catch PKCS#1 and unencrypted PKCS#8
+	 * (except in OpenSSL 0.9.8 where it doesn't handle
+	 * the latter but nobody cares about 0.9.8 any more. */
+	key = d2i_PrivateKey_fp(f, NULL);
+	if (key) {
+		ret = 0;
+		if (!SSL_CTX_use_PrivateKey(vpninfo->https_ctx, key)) {
+			vpn_progress(vpninfo, PRG_ERR,
+				     _("Loading private key failed\n"));
+			openconnect_report_ssl_errors(vpninfo);
+			ret = -EINVAL;
+		}
+		EVP_PKEY_free(key);
+		fclose(f);
+		return ret;
+	} else {
+		/* Encrypted PKCS#8 DER */
+		X509_SIG *p8;
+
+		fseek(f, 0, SEEK_SET);
+		p8 = d2i_PKCS8_fp(f, NULL);
+
+		if (p8) {
+			PKCS8_PRIV_KEY_INFO *p8inf;
+			char *pass = vpninfo->cert_password;
+
+			fclose(f);
+
+			while (!(p8inf = PKCS8_decrypt(p8, pass ? : "", pass ? strlen(pass) : 0))) {
+				unsigned long err = ERR_peek_error();
+
+				if (ERR_GET_LIB(err) == ERR_LIB_EVP &&
+				    ERR_GET_FUNC(err) == EVP_F_EVP_DECRYPTFINAL_EX &&
+				    ERR_GET_REASON(err) == EVP_R_BAD_DECRYPT) {
+					ERR_clear_error();
+
+					if (pass) {
+						vpn_progress(vpninfo, PRG_ERR,
+							     _("Failed to decrypt PKCS#8 certificate file\n"));
+						free(pass);
+						pass = NULL;
+					}
+
+					if (request_passphrase(vpninfo, "openconnect_pkcs8", &pass,
+							       _("Enter PKCS#8 pass phrase:")) >= 0)
+						continue;
+				} else {
+					vpn_progress(vpninfo, PRG_ERR,
+						     _("Failed to decrypt PKCS#8 certificate file\n"));
+					openconnect_report_ssl_errors(vpninfo);
+				}
+
+				free(pass);
+				vpninfo->cert_password = NULL;
+
+				X509_SIG_free(p8);
+				return -EINVAL;
+			}
+			free(pass);
+			vpninfo->cert_password = NULL;
+
+			key = EVP_PKCS82PKEY(p8inf);
+
+			PKCS8_PRIV_KEY_INFO_free(p8inf);
+			X509_SIG_free(p8);
+
+			if (key == NULL) {
+				vpn_progress(vpninfo, PRG_ERR,
+					     _("Failed to convert PKCS#8 to OpenSSL EVP_PKEY\n"));
+				return -EIO;
+			}
+			ret = 0;
+			if (!SSL_CTX_use_PrivateKey(vpninfo->https_ctx, key)) {
+				vpn_progress(vpninfo, PRG_ERR,
+					     _("Loading private key failed\n"));
+				openconnect_report_ssl_errors(vpninfo);
+				ret = -EINVAL;
+			}
+			EVP_PKEY_free(key);
+			return ret;
+		}
+	}
+
+	fclose(f);
 	vpn_progress(vpninfo, PRG_ERR,
 		     _("Failed to identify private key type in '%s'\n"),
 		     vpninfo->sslkey);
@@ -957,11 +1099,11 @@ int get_cert_md5_fingerprint(struct openconnect_info *vpninfo,
 
 static int set_peer_cert_hash(struct openconnect_info *vpninfo)
 {
-	unsigned char sha1[SHA1_SIZE];
+	unsigned char sha256_hash[SHA256_SIZE];
+	unsigned char sha1_hash[SHA1_SIZE];
 	EVP_PKEY *pkey;
 	BIO *bp = BIO_new(BIO_s_mem());
 	BUF_MEM *keyinfo;
-	int i;
 
 	/* We can't use X509_pubkey_digest() because it only hashes the
 	   subjectPublicKey BIT STRING, and not the whole of the
@@ -977,21 +1119,18 @@ static int set_peer_cert_hash(struct openconnect_info *vpninfo)
 
 	BIO_get_mem_ptr(bp, &keyinfo);
 
-	openconnect_sha1(sha1, keyinfo->data, keyinfo->length);
+	openconnect_sha256(sha256_hash, keyinfo->data, keyinfo->length);
+	openconnect_sha1(sha1_hash, keyinfo->data, keyinfo->length);
 
 	BIO_free(bp);
 
-	vpninfo->peer_cert_hash = malloc(SHA1_SIZE * 2 + 6);
-	if (vpninfo->peer_cert_hash) {
-		snprintf(vpninfo->peer_cert_hash, 6, "sha1:");
-		for (i = 0; i < sizeof(sha1); i++)
-			sprintf(&vpninfo->peer_cert_hash[i*2 + 5], "%02x", sha1[i]);
-	}
+	vpninfo->peer_cert_sha1 = openconnect_bin2hex("sha1:", sha1_hash, sizeof(sha1_hash));
+	vpninfo->peer_cert_sha256 = openconnect_bin2hex("sha256:", sha256_hash, sizeof(sha256_hash));
 
 	return 0;
 }
 
-#if OPENSSL_VERSION_NUMBER < 0x10002000L
+#if OPENSSL_VERSION_NUMBER < 0x10002000L || defined(LIBRESSL_VERSION_NUMBER)
 static int match_hostname_elem(const char *hostname, int helem_len,
 			       const char *match, int melem_len)
 {
@@ -1265,7 +1404,8 @@ static int match_cert_hostname(struct openconnect_info *vpninfo, X509 *peer_cert
 			else
 				family = AF_INET6;
 
-			inet_ntop(family, ipaddr, host, sizeof(host));
+			/* In Windows, the 'src' argument of inet_ntop() isn't const */
+			inet_ntop(family, (void *)ipaddr, host, sizeof(host));
 			vpn_progress(vpninfo, PRG_DEBUG,
 				     _("Matched %s address '%s'\n"),
 				     (family == AF_INET6) ? "IPv6" : "IPv4",
@@ -1300,18 +1440,16 @@ static int match_cert_hostname(struct openconnect_info *vpninfo, X509 *peer_cert
 static void workaround_openssl_certchain_bug(struct openconnect_info *vpninfo,
 					     SSL *ssl)
 {
-	/* OpenSSL has problems with certificate chains — if there are
+	/* OpenSSL has problems with certificate chains -- if there are
 	   multiple certs with the same name, it doesn't necessarily
-	   choose the _right_ one. (RT#1942). This affects some servers,
-	   and we can work around it by picking the right ones ourselves
-	   and adding them manually on the wire. */
+	   choose the _right_ one. (RT#1942)
+	   Pick the right ones for ourselves and add them manually. */
 	X509 *cert = SSL_get_certificate(ssl);
 	X509 *cert2;
 	X509_STORE *store = SSL_CTX_get_cert_store(vpninfo->https_ctx);
 	X509_STORE_CTX *ctx;
 	void *extra_certs;
-	STACK_OF(X509) *chain;
-	int i;
+	X509_STORE_CTX_get_issuer_fn issuer_fn;
 
 	if (!cert || !store)
 		return;
@@ -1327,16 +1465,10 @@ static void workaround_openssl_certchain_bug(struct openconnect_info *vpninfo,
 	if (X509_STORE_CTX_init(ctx, store, NULL, NULL))
 		goto out;
 
-	X509_STORE_CTX_set_cert(ctx, cert);
+	issuer_fn = X509_STORE_CTX_get_get_issuer(ctx);
 
-	/* We only really want it to build the chain; we don't really
-	   care if it's trusted */
-	(void)X509_verify_cert(ctx);
-
-	chain = X509_STORE_CTX_get0_chain(ctx);
-	for (i = 1; i < sk_X509_num(chain); i++) {
+	while (issuer_fn(&cert2, ctx, cert) == 1) {
 		char buf[200];
-		cert2 = sk_X509_value(chain, i);
 		if (cert2 == cert)
 			break;
 		if (X509_check_issued(cert2, cert2) == X509_V_OK)
@@ -1346,10 +1478,9 @@ static void workaround_openssl_certchain_bug(struct openconnect_info *vpninfo,
 				  buf, sizeof(buf));
 		vpn_progress(vpninfo, PRG_DEBUG,
 			     _("Extra cert from cafile: '%s'\n"), buf);
-		X509_up_ref(cert);
 		SSL_CTX_add_extra_chain_cert(vpninfo->https_ctx, cert);
 	}
-out:
+ out:
 	X509_STORE_CTX_free(ctx);
 }
 
@@ -1401,6 +1532,9 @@ static int ssl_app_verify_callback(X509_STORE_CTX *ctx, void *arg)
 	struct openconnect_info *vpninfo = arg;
 	const char *err_string = NULL;
 	X509 *cert = X509_STORE_CTX_get0_cert(ctx);
+#ifdef X509_V_FLAG_PARTIAL_CHAIN
+	X509_VERIFY_PARAM *param;
+#endif
 
 	if (vpninfo->peer_cert) {
 		/* This is a *rehandshake*. Require that the server
@@ -1418,6 +1552,11 @@ static int ssl_app_verify_callback(X509_STORE_CTX *ctx, void *arg)
 
 	set_peer_cert_hash(vpninfo);
 
+#ifdef X509_V_FLAG_PARTIAL_CHAIN
+	param = X509_STORE_CTX_get0_param(ctx);
+	if (param)
+		X509_VERIFY_PARAM_set_flags(param, X509_V_FLAG_PARTIAL_CHAIN);
+#endif
 	if (!X509_verify_cert(ctx)) {
 		err_string = X509_verify_cert_error_string(X509_STORE_CTX_get_error(ctx));
 	} else {
@@ -1463,7 +1602,7 @@ static int ssl_app_verify_callback(X509_STORE_CTX *ctx, void *arg)
 
 static int check_certificate_expiry(struct openconnect_info *vpninfo)
 {
-	ASN1_TIME *notAfter;
+	method_const ASN1_TIME *notAfter;
 	const char *reason = NULL;
 	time_t t;
 	int i;
@@ -1472,7 +1611,7 @@ static int check_certificate_expiry(struct openconnect_info *vpninfo)
 		return 0;
 
 	t = time(NULL);
-	notAfter = X509_get_notAfter(vpninfo->cert_x509);
+	notAfter = X509_get0_notAfter(vpninfo->cert_x509);
 	i = X509_cmp_time(notAfter, &t);
 	if (!i) {
 		vpn_progress(vpninfo, PRG_ERR,
@@ -1507,7 +1646,6 @@ static int check_certificate_expiry(struct openconnect_info *vpninfo)
 
 int openconnect_open_https(struct openconnect_info *vpninfo)
 {
-	method_const SSL_METHOD *ssl3_method;
 	SSL *https_ssl;
 	BIO *https_bio;
 	int ssl_sock;
@@ -1520,26 +1658,53 @@ int openconnect_open_https(struct openconnect_info *vpninfo)
 		X509_free(vpninfo->peer_cert);
 		vpninfo->peer_cert = NULL;
 	}
-	free (vpninfo->peer_cert_hash);
-	vpninfo->peer_cert_hash = NULL;
+	free (vpninfo->peer_cert_sha1);
+	vpninfo->peer_cert_sha1 = NULL;
+	free (vpninfo->peer_cert_sha256);
+	vpninfo->peer_cert_sha256 = NULL;
 	vpninfo->cstp_cipher = NULL;
 
 	ssl_sock = connect_https_socket(vpninfo);
 	if (ssl_sock < 0)
 		return ssl_sock;
 
-	ssl3_method = TLSv1_client_method();
 	if (!vpninfo->https_ctx) {
-		vpninfo->https_ctx = SSL_CTX_new(ssl3_method);
+#if OPENSSL_VERSION_NUMBER < 0x10100000L || defined(LIBRESSL_VERSION_NUMBER)
+		vpninfo->https_ctx = SSL_CTX_new(SSLv23_client_method());
+		if (vpninfo->https_ctx)
+			SSL_CTX_set_options(vpninfo->https_ctx, SSL_OP_NO_SSLv2|SSL_OP_NO_SSLv3);
+#else
+		vpninfo->https_ctx = SSL_CTX_new(TLS_client_method());
+		if (vpninfo->https_ctx &&
+		    !SSL_CTX_set_min_proto_version(vpninfo->https_ctx, TLS1_VERSION)) {
+			SSL_CTX_free(vpninfo->https_ctx);
+			vpninfo->https_ctx = NULL;
+		}
+#endif
+		if (!vpninfo->https_ctx) {
+			vpn_progress(vpninfo, PRG_ERR,
+				     _("Create TLSv1 CTX failed\n"));
+			openconnect_report_ssl_errors(vpninfo);
+			return -EINVAL;
+		}
 
-		/* Some servers (or their firewalls) really don't like seeing
-		   extensions. */
-#ifdef SSL_OP_NO_TICKET
+		/* Try to work around the broken firewalls which reject ClientHello
+		 * packets in certain size ranges. If we have SSL_OP_TLSEXT_PADDING
+		 * use it, else fall back to SSL_OP_NO_TICKET which mostly worked for
+		 * a long time. */
+#if defined(SSL_OP_TLSEXT_PADDING)
+		SSL_CTX_set_options(vpninfo->https_ctx, SSL_OP_TLSEXT_PADDING);
+#elif defined(SSL_OP_NO_TICKET)
 		SSL_CTX_set_options(vpninfo->https_ctx, SSL_OP_NO_TICKET);
 #endif
 
 		if (vpninfo->cert) {
 			err = load_certificate(vpninfo);
+			if (!err && !SSL_CTX_check_private_key(vpninfo->https_ctx)) {
+				vpn_progress(vpninfo, PRG_ERR,
+					     _("SSL certificate and key do not match\n"));
+				err = -EINVAL;
+			}
 			if (err) {
 				vpn_progress(vpninfo, PRG_ERR,
 					     _("Loading certificate failed. Aborting.\n"));
@@ -1653,7 +1818,7 @@ int openconnect_open_https(struct openconnect_info *vpninfo)
 	 * 4fcdd66fff5fea0cfa1055c6680a76a4303f28a2
 	 * cd6bd5ffda616822b52104fee0c4c7d623fd4f53
 	 */
-#if OPENSSL_VERSION_NUMBER >= 0x10001070
+#if OPENSSL_VERSION_NUMBER >= 0x10001070 && !defined(LIBRESSL_VERSION_NUMBER)
 	if (string_is_hostname(vpninfo->hostname))
 		SSL_set_tlsext_host_name(https_ssl, vpninfo->hostname);
 #endif
