@@ -241,7 +241,7 @@ int connect_https_socket(struct openconnect_info *vpninfo)
 		   different types of returned sockaddr_in{6,}. */
 #ifdef LIBPROXY_HDR
 		if (vpninfo->proxy_factory) {
-			char *url;
+			struct oc_text_buf *url_buf = buf_alloc();
 			char **proxies;
 			int i = 0;
 
@@ -250,20 +250,18 @@ int connect_https_socket(struct openconnect_info *vpninfo)
 			free(vpninfo->proxy);
 			vpninfo->proxy = NULL;
 
-			if (vpninfo->port == 443)
-				i = asprintf(&url, "https://%s/%s", vpninfo->hostname,
-					     vpninfo->urlpath?:"");
-			else
-				i = asprintf(&url, "https://%s:%d/%s", vpninfo->hostname,
-					     vpninfo->port, vpninfo->urlpath?:"");
-			if (i == -1) {
+			buf_append(url_buf, "https://%s", vpninfo->hostname);
+			if (vpninfo->port != 443)
+				buf_append(url_buf, ":%d", vpninfo->port);
+			buf_append(url_buf, "/%s", vpninfo->urlpath?:"");
+			if (buf_error(url_buf)) {
+				buf_free(url_buf);
 				ssl_sock = -ENOMEM;
 				goto out;
 			}
 
 			proxies = px_proxy_factory_get_proxies(vpninfo->proxy_factory,
-							       url);
-
+							       url_buf->data);
 			i = 0;
 			while (proxies && proxies[i]) {
 				if (!vpninfo->proxy &&
@@ -275,7 +273,7 @@ int connect_https_socket(struct openconnect_info *vpninfo)
 						  NULL, 0);
 				i++;
 			}
-			free(url);
+			buf_free(url_buf);
 			free(proxies);
 			if (vpninfo->proxy)
 				vpn_progress(vpninfo, PRG_DEBUG,
@@ -615,7 +613,7 @@ int openconnect_passphrase_from_fsid(struct openconnect_info *vpninfo)
 }
 #endif
 
-#if defined(OPENCONNECT_OPENSSL) || defined(DTLS_OPENSSL)
+#if defined(OPENCONNECT_OPENSSL)
 /* We put this here rather than in openssl.c because it might be needed
    for OpenSSL DTLS support even when GnuTLS is being used for HTTPS */
 int openconnect_print_err_cb(const char *str, size_t len, void *ptr)
@@ -910,9 +908,13 @@ int udp_sockaddr(struct openconnect_info *vpninfo, int port)
 	if (vpninfo->peer_addr->sa_family == AF_INET) {
 		struct sockaddr_in *sin = (void *)vpninfo->dtls_addr;
 		sin->sin_port = htons(port);
+		vpninfo->dtls_tos_proto = IPPROTO_IP;
+		vpninfo->dtls_tos_optname = IP_TOS;
 	} else if (vpninfo->peer_addr->sa_family == AF_INET6) {
 		struct sockaddr_in6 *sin = (void *)vpninfo->dtls_addr;
 		sin->sin6_port = htons(port);
+		vpninfo->dtls_tos_proto = IPPROTO_IPV6;
+		vpninfo->dtls_tos_optname = IPV6_TCLASS;
 	} else {
 		vpn_progress(vpninfo, PRG_ERR,
 			     _("Unknown protocol family %d. Cannot create UDP server address\n"),
@@ -920,6 +922,11 @@ int udp_sockaddr(struct openconnect_info *vpninfo, int port)
 		return -EINVAL;
 	}
 
+	/* in case DTLS TOS copy is disabled, reset the optname value */
+	/* so that the copy won't be applied in dtls.c / dtls_mainloop() */
+	if (!vpninfo->dtls_pass_tos)
+		vpninfo->dtls_tos_optname = 0;
+
 	return 0;
 }
 
